/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 *  PrimeBase Media Stream (PBMS)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2008-09-10	Barry Leslie
 *
 * H&G2JCtL
 */
#include "CSConfig.h"

#include <curl/curl.h>

#ifdef DRIZZLED
#include "libdrizzle/drizzle.h"
#define MYSQL DRIZZLE
#define MYSQL_RES DRIZZLE_RES
#define MYSQL_ROW DRIZZLE_ROW

#define mysql_query			drizzle_query
#define mysql_store_result	drizzle_store_result
#define mysql_errno			drizzle_errno
#define mysql_error			drizzle_error
#define mysql_num_rows		drizzle_num_rows
#define mysql_fetch_row		drizzle_fetch_row
#define mysql_free_result	drizzle_free_result

#else
#include "mysql.h"
#endif

#include "pbmslib.h"
#include "pbms.h"
#include "CSGlobal.h"
#include "CSThread.h"
#include "CSString.h"
#include "CSStrUtil.h"
//#include "CSSocket.h"
#include "CSHTTPStream.h"
#include "CSMd5.h"
#include "Util_ms.h"
#include "metadata_ms.h"

#define CLEAR_SELF()	CSThread::setSelf(NULL)
#define MAX_STMT_SIZE	1024

static int global_errno;
static char global_err_message[MS_RESULT_MESSAGE_SIZE];

static unsigned long init_count = 0;
static CSThreadList	*pbms_thread_list = NULL;
static	CSThread *mslib_global_thread = NULL;

static void report_global_mse_error(CSThread *thd)
{
	global_errno = thd->myException.getErrorCode();
	cs_strcpy(MS_RESULT_MESSAGE_SIZE, global_err_message,  thd->myException.getMessage());
}

#define DFLT_CONNECTION_TIMEOUT	10	// Changing this required a documentation update.

#define THROW_CURL_IF(v) { if (v) CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, ms_curl_error);}
//======================================
static size_t receive_data(void *ptr, size_t size, size_t nmemb, void *stream);
static size_t receive_header(void *ptr, size_t size, size_t nmemb, void *stream);
static size_t send_callback(void *ptr, size_t size, size_t nmemb, void *stream);
//--------------------------------------------
class  PBMS_ConHandle:public CSThread {
	public:
	
	// Fields that must be freed when the object is destroyed.
	CSString			*ms_host;
	CSString			*ms_database;
	struct curl_slist	*ms_header_list;	// A curl list of headers to be sent with the next PUT.
	struct curl_slist	*ms_info_only;	// A curl list of headers to be sent with Get.
	struct curl_slist	*ms_ping_header;	// A curl list of headers to be sent with ping.
	CSStringBuffer		*ms_url_str;
	CURL				*ms_curl;	
	CSStringBuffer		*ms_buffer;
	CSStringBuffer		*ms_errorReply;

	PBMS_ConHandle():
		CSThread(pbms_thread_list),
		ms_host(NULL),
		ms_database(NULL),
		ms_port(0),
		ms_transmition_timeout(0),
		ms_curl(NULL),
		ms_header_list(NULL),
		ms_info_only(NULL),
		ms_ping_header(NULL),
		ms_buffer(NULL),
		ms_errorReply(NULL),
		ms_url_str(NULL),
		ms_errno(0),
		ms_throw_error(false)
	{
		}
		
	~PBMS_ConHandle() {
		if (ms_curl) 
			curl_easy_cleanup(ms_curl);
			
		if (ms_database) 
			ms_database->release();
			
		if (ms_host) 
			ms_host->release();
			
		if (ms_buffer) 
			ms_buffer->release();
			
		if (ms_errorReply) 
			ms_errorReply->release();
			
		if (ms_url_str) 
			ms_url_str->release();
			
		if (ms_header_list) 
			curl_slist_free_all(ms_header_list);
			
		if (ms_info_only) 
			curl_slist_free_all(ms_info_only);
			
		if(ms_ping_header)
			curl_slist_free_all(ms_ping_header);
			
		ms_headers.clearHeaders();
		ms_metadata_out.clearHeaders();
			
	}
	unsigned int		ms_replyStatus;
	CSHTTPHeaders		ms_headers; 
	CSHTTPHeaders		ms_metadata_out;
	u_int				ms_next_header;
	u_int				ms_max_header;	
	unsigned int		ms_port;
	unsigned int		ms_transmition_timeout; // In the future this may have some effect but for now it is always be 0 (no timeout).
	unsigned int		ms_url_base_len;
	bool				ms_throw_error;	// Gets set if an exception occurs in a callback.

	int					ms_errno;
	char				ms_err_message[MS_RESULT_MESSAGE_SIZE];
	
	
	char ms_curl_error[CURL_ERROR_SIZE];
	
	size_t ms_DataToBeTransfered;
	// Get data caller parameters:
	char				*ms_getBuffer;
	size_t				ms_getBufferSize;
	size_t				ms_getDataLen;
	PBMS_CALLBACK_FUNC	ms_getCB;
	void				*ms_getCBData;
	bool				ms_getCBCalled;
	
	void set_get_user_data(char *buffer, size_t buffer_size, PBMS_CALLBACK_FUNC cb = NULL, void *caller_data = NULL)	
	{
		ms_getDataLen = 0;
		ms_getBuffer = buffer;
		ms_getBufferSize = buffer_size;
		ms_getCB = cb;
		ms_getCBData = caller_data;
		ms_getCBCalled = false;
		
	}
	
	// Put data caller parameters:
	char				*ms_putData;
	size_t				ms_putDataLen;
	size_t				ms_putDataOffset;
	PBMS_CALLBACK_FUNC	ms_putCB;
	void				*ms_putCBData;
	bool				ms_putCBCalled;
	
	void set_put_user_data(char *buffer, size_t buffer_size, size_t expected_data_size, PBMS_CALLBACK_FUNC cb = NULL, void *caller_data = NULL)	
	{
		ms_DataToBeTransfered = expected_data_size;
		ms_putData = buffer;
		ms_putDataLen = buffer_size;
		ms_putDataOffset =0;
		ms_putCB = cb;
		ms_putCBData = caller_data;
		ms_putCBCalled = false;
		
	}
	
	
	void ms_initConnection(const char* host, unsigned int port, const char* database)
	{
		ms_curl = curl_easy_init();
		if (!ms_curl)
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_easy_init() failed.");
		
		if (curl_easy_setopt(ms_curl, CURLOPT_ERRORBUFFER, ms_curl_error))
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_easy_setopt(CURLOPT_ERRORBUFFER) failed.");
		
		// Uncomment this line to trace network action during request. Very Usefull!!
		//curl_easy_setopt(ms_curl, CURLOPT_VERBOSE, 1L);
		
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_TCP_NODELAY, 1L));

		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_NOPROGRESS, 1L));
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_WRITEFUNCTION, receive_data));
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_READFUNCTION, send_callback));	
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HEADERFUNCTION, receive_header));
		
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_WRITEDATA, this));
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_READDATA, this));
		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_WRITEHEADER, this));

		THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_FOLLOWLOCATION, 1L)); // Follow redirects.

		ms_port = port;
		pbms_api_owner = true;
		ms_host = CSString::newString(host);
		ms_database = CSString::newString(database);
		
		ms_url_str = new CSStringBuffer(50);	
		ms_url_str->append("http://");
		ms_url_str->append(host);
		ms_url_str->append(":");
		ms_url_str->append((int)port);
		ms_url_str->append("/");
		ms_url_base_len	= 	ms_url_str->length();
			
		ms_info_only = curl_slist_append(ms_info_only, MS_BLOB_INFO_REQUEST ": yes");
		if (!ms_info_only) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");
			
		ms_ping_header = curl_slist_append(ms_ping_header, MS_PING_REQUEST ": yes");
		if (!ms_ping_header) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");


		ms_buffer = new CSStringBuffer(50);	
	}
	
	void ms_ping();
	
	void ms_setSelf()
	{
		setSelf(this);
	}

	void ms_init_put_blob(curl_off_t size, const char *table, const char *alias, const char *checksum);
	
	void ms_init_get_blob(const char *ref, bool is_alias, bool info_only);
	
	void ms_get_info(const char *ref, bool is_alias);
	
	u_int ms_init_fetch() {ms_next_header =0; return ms_max_header = ms_headers.numHeaders();}
	
	bool ms_next(const char **name, const char **value) 
	{
		if (ms_next_header >= ms_max_header)
			return false;
			
		CSHeader *header = ms_headers.getHeader(ms_next_header++);
		*name = header->getNameCString();
		*value = header->getValueCString();
		header->release();
		return true;
	}
	
	void dump_headers() 
	{
		u_int i = 0;
		CSHeader *header;
		printf("Headers:\n");
		printf("---------------------------------------\n");
		while  ( (header = ms_headers.getHeader(i++)) ) {
			printf("%s : %s\n", header->getNameCString(), header->getValueCString());
			header->release();			
		}
		printf("---------------------------------------\n");
	}
					
	CSString	*ms_get_metadata(const char *name) 
	{ 
		return ms_headers.getHeaderValue(name);
	}
					
	CSString	*ms_get_alias() {return ms_get_metadata(MS_ALIAS_TAG);}

	CSString	*ms_get_checksum() {return ms_get_metadata(MS_CHECKSUM_TAG);}

	void report_error(CSThread *self)
	{
		ms_errno = self->myException.getErrorCode();
		cs_strcpy(MS_RESULT_MESSAGE_SIZE, ms_err_message,  self->myException.getMessage());
	}

	void throw_http_reply_exception();
	
	void check_reply_status() 
	{
		switch (ms_replyStatus) {
			case 200:
			//case 301: // Moved Permanently
			//case 307: // Temporary Redirect
				break;
			default:
				throw_http_reply_exception();
		}
		
	}
	
};

// CURL callback functions:
////////////////////////////
static size_t receive_data(void *vptr, size_t objs, size_t obj_size, void *v_con)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) v_con;
	size_t data_len = objs * obj_size;
	char *ptr = (char*)vptr;

	if (con->ms_replyStatus >= 400) { // Collect the error reply.
		enter_();
		try_(a) {
			if (!con->ms_errorReply)
				con->ms_errorReply = new CSStringBuffer(50);		
			con->ms_errorReply->append(ptr, data_len);
		}
		catch_(a);
		con->ms_throw_error = true;
		data_len = -1;
		
		cont_(a);
		return_(data_len);
	}
	
	do {
		size_t copy_len = 0;
		
		if (data_len > con->ms_DataToBeTransfered) { // This should never happen.
			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Unexpected data received.");
			con->ms_throw_error = true;
			return-1;
		}
		// Copy as much data into the callers buffer as there is space for it.
		if (con->ms_getBufferSize) {
				
			if (data_len > con->ms_getBufferSize) 
				copy_len = con->ms_getBufferSize;
			else
				copy_len = data_len;

			memcpy(con->ms_getBuffer, ptr, copy_len);
			data_len -= copy_len;
			ptr += copy_len;
			con->ms_getDataLen += copy_len;
			con->ms_DataToBeTransfered -= copy_len;
		}
		
		// Call the callback with the current buffer of data, if there is one, and get the next buffer to fill.
		if (con->ms_getCB) {			
			con->ms_getCBCalled = true; // API specs guarentee that the callback gets called at least once.
			con->ms_getBufferSize = copy_len; // The callback will reset this to the buffer size if required.
			con->ms_getBuffer =  (con->ms_getCB)(con->ms_getCBData, con->ms_getBuffer, &con->ms_getBufferSize, con->ms_DataToBeTransfered);
		} else if (!data_len) {
			con->ms_getBuffer += copy_len;
		} else
			break;
			
	} while (con->ms_getBuffer && data_len);
	
	if (data_len) { // More data received than expected.
		CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob data overflow.");
		con->ms_throw_error = true;
		return-1;
	}
	
	return objs * obj_size; // All the data was received.
}

#define IS_REDIRECT(s) ((s >= 300) && (s < 400))
//----------------------
static size_t receive_header(void *header, size_t objs, size_t obj_size, void *v_con)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) v_con;
	size_t size = objs * obj_size;
	char *end, *ptr = (char*) header, *name, *value;
	u_int name_len, value_len;
	
	end = ptr + size;
	if (*(end -2) == '\r' && *(end -1) == '\n')
		end -=2;
		
	while ((end != ptr) && (*ptr == ' ')) ptr++;
	if (end == ptr)
		return size;
	
	// Get the reply status.	
	if (((!con->ms_replyStatus) || IS_REDIRECT(con->ms_replyStatus) ) && !strncasecmp(ptr, "HTTP", 4)) {
		char status[4];
		while ((end != ptr) && (*ptr != ' ')) ptr++; // skip HTTP stuff
		while ((end != ptr) && (*ptr == ' ')) ptr++; // find the start of eh status code.
		if (end == ptr)
			return size;
			
		if (end < (ptr +3)) // expecting a 3 digit status code.
			return size;
			
		memcpy(status, ptr, 3);
		status[3] = 0;
		
		con->ms_replyStatus = atoi(status);
	}
	
	name = ptr;
	while ((end != ptr) && (*ptr != ':')) ptr++;
	if (end == ptr)
		return size;
	name_len = ptr - name;
	
	ptr++; 
	while ((end != ptr) && (*ptr == ' ')) ptr++;
	if (end == ptr)
		return size;
	
	value = ptr;
	value_len = end - value;
	
	while (name[name_len-1] == ' ') name_len--;
	while (value[value_len-1] == ' ') value_len--;
	
	if (!strncasecmp(name, "Content-Length", 14)) {
		char len[32];
		memcpy(len, value, (value_len > 31)?31:value_len);
		len[(value_len > 31)?31:value_len] = 0;
		con->ms_DataToBeTransfered = atoll(len);
	}
	
	enter_();
	try_(a) {
		con->ms_headers.addHeader(name, name_len, value, value_len);
	}
	
	catch_(a);
	con->ms_throw_error = true;
	return_(-1);
		
	cont_(a);
	return_(size);
}

//----------------------
static size_t send_callback(void *ptr, size_t objs, size_t obj_size, void *v_con)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) v_con;
	char *buffer = (char*) ptr;
	size_t data_sent = 0, data_size, buffer_size = objs * obj_size;

	do {
		// Copy as much data into the curl buffer as there is space for it.
		data_size = con->ms_putDataLen - con->ms_putDataOffset;
		if (data_size) {
			size_t copy_len;
			if (data_size > buffer_size)
				copy_len = buffer_size;
			else 
				copy_len = data_size;
				
			memcpy(buffer, con->ms_putData + con->ms_putDataOffset, copy_len);
			buffer += copy_len;
			buffer_size -= copy_len;
			con->ms_putDataOffset += copy_len;
			data_sent += copy_len;
			con->ms_DataToBeTransfered -= copy_len;
		}
		
		// If there is still buffer space get the next block of data.
		if (con->ms_putCB && buffer_size) {			
			con->ms_putDataOffset = 0;
			//con->ms_putDataLen = 0; The callback expects this to be the size of the last block sent.
			con->ms_putCBCalled = true; // API specs guarentee that the callback gets called at least once.
			con->ms_putData =  (con->ms_putCB)(con->ms_putCBData, con->ms_putData, &con->ms_putDataLen, con->ms_DataToBeTransfered);
			if (!con->ms_putData)
				con->ms_putDataLen = 0;
		} else if (buffer_size)
			break;
			
	} while (con->ms_putData && buffer_size);
	
	if (buffer_size && con->ms_DataToBeTransfered) { // Less data sent than expected.
		CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Blob smaller than expected.");
		con->ms_throw_error = true;
		return-1;
	}
	 if (!con->ms_putData) {
		con->ms_putCB = NULL;
	}
	return data_sent;
}

#define CONTENT_TYPE "Content-Type"
//------------------------------------------------
void PBMS_ConHandle::ms_init_put_blob(curl_off_t size, const char *table, const char *alias, const char *checksum)
{
	char buffer[MS_META_VALUE_SIZE + MS_META_NAME_SIZE +2];
	int buffer_size = MS_META_VALUE_SIZE + MS_META_NAME_SIZE +2;
	bool have_content_type = false;
	
	ms_url_str->setLength(ms_url_base_len);

	ms_url_str->append(ms_database->getCString());
	if (table) {
		ms_url_str->append("/");
		ms_url_str->append(table);
	}
	
	// Remove any old headers
	if (ms_header_list) {
		curl_slist_free_all(ms_header_list);
		ms_header_list = NULL;
	}
	
	// Add metadata headers.
	u_int i = 0;
	CSHeader *header;
	while  ( (header = ms_metadata_out.getHeader(i++)) ) {
		cs_strcpy(buffer_size, buffer, header->getNameCString());
		if (!strcasecmp(  buffer, CONTENT_TYPE))
			have_content_type = true;
		cs_strcat(buffer_size, buffer, ':');
		cs_strcat(buffer_size, buffer, header->getValueCString());
		header->release();			
		ms_header_list = curl_slist_append(ms_header_list, buffer);
		if (!ms_header_list) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");
	}
	
	if (!have_content_type) {
		// Prevent CURLOPT_POST from adding a content type. 
		ms_header_list = curl_slist_append(ms_header_list, CONTENT_TYPE ":");
		if (!ms_header_list) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");
	}
		
	if (checksum) {	
		cs_strcpy(buffer_size, buffer, MS_CHECKSUM_TAG ":");
		cs_strcat(buffer_size, buffer, checksum);
		
		ms_header_list = curl_slist_append(ms_header_list, buffer);
		if (!ms_header_list) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");
	}
		
	
	if (alias) {
		if (strlen(alias) > MS_META_VALUE_SIZE) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "BLOB alias name too long.");

		cs_strcpy(buffer_size, buffer, MS_ALIAS_TAG ":");
		cs_strcat(buffer_size, buffer, alias);
		
		ms_header_list = curl_slist_append(ms_header_list, buffer);
		if (!ms_header_list) 
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "curl_slist_append() failed.");
	}
	
	
	// Using CURLOPT_UPLOAD is VERY slow!
	//THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_INFILESIZE_LARGE, size));
	//THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_UPLOAD, 1L));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_POSTFIELDSIZE_LARGE, size));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_POST, 1L));

	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPHEADER, ms_header_list));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_URL, ms_url_str->getCString() ));	
		
	ms_replyStatus = 0;
	ms_headers.clearHeaders();
	if (ms_errorReply)
		ms_errorReply->setLength(0);
}
	

//------------------------------------------------
void PBMS_ConHandle::ms_init_get_blob(const char *ref, bool is_alias, bool info_only)
{
	MSBlobURLRec blob;
	
	ms_url_str->setLength(ms_url_base_len);
	
	if (is_alias || !ms_parse_blob_url(&blob, ref)) {
		ms_url_str->append(ms_database->getCString());
		ms_url_str->append("/");
	}
	ms_url_str->append((char *) ref);
	
	//THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_UPLOAD, 0L));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_POST, 0L));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPHEADER, (info_only)?ms_info_only: NULL));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_URL, ms_url_str->getCString()));
	
	ms_replyStatus = 0;
	ms_headers.clearHeaders();
	if (ms_errorReply)
		ms_errorReply->setLength(0);
}
	
//------------------------------------------------
void PBMS_ConHandle::ms_get_info(const char *ref, bool is_alias)
{
	enter_();
	ms_init_get_blob(ref, is_alias, true);
	
	if (curl_easy_perform(ms_curl) && !ms_throw_error)
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, ms_curl_error);
	
	if (ms_throw_error)
		throw_();
		
	check_reply_status();
	
	exit_();
}

//------------------------------------------------
void PBMS_ConHandle::ms_ping()
{
	enter_();
	ms_url_str->setLength(ms_url_base_len);
	
//	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_UPLOAD, 0L));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_POST, 0L));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_HTTPHEADER, ms_ping_header));
	THROW_CURL_IF(curl_easy_setopt(ms_curl, CURLOPT_URL, ms_url_str->getCString()));
	
	ms_replyStatus = 0;
	ms_headers.clearHeaders();
	if (ms_errorReply)
		ms_errorReply->setLength(0);

	if (curl_easy_perform(ms_curl) && !ms_throw_error)
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, ms_curl_error);
	
	if (ms_throw_error)
		throw_();
		
	check_reply_status();
	
	exit_();
}

//------------------------------------------------
void PBMS_ConHandle::throw_http_reply_exception()
{
	CSString *reply = NULL, *error_text = NULL;
	
	enter_();
	
	try_(a) {
		size_t size = 0;
		 //dump_headers();
		 
		if (ms_errorReply)
			size = ms_errorReply->length();
		
		if (!size) {
			error_text = CSString::newString("Missing HTTP reply: possible Media Stream engine connection failure.");
		} else {
			u_int start, end;
		
			reply = CSString::newString(ms_errorReply);
			ms_errorReply = NULL;
			
			start = reply->locate(EXCEPTION_REPLY_MESSAGE_PREFIX_TAG, 1);
			start += strlen(EXCEPTION_REPLY_MESSAGE_PREFIX_TAG);
			end = reply->locate(EXCEPTION_REPLY_MESSAGE_SUFFIX_TAG, 1); 
			if (start < end)
				error_text = reply->substr(start, end - start);
			else {
				error_text = reply;
				reply->retain();
			}
		}
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, error_text->getCString());
	}
	
	finally_(a) {
		if (reply) reply->release();
		if (error_text) error_text->release();
	}	
	cont_(a);
	
	exit_();
}

//------------------------------------------------
static void report_global_error(int err, const char *message, int line)
{
	char line_no[32];
	
	if (global_errno)
		return;
	global_errno = err;
	cs_strcpy(MS_RESULT_MESSAGE_SIZE, global_err_message,  message);
	sprintf(line_no, ": line %d", line);
	cs_strcat(MS_RESULT_MESSAGE_SIZE, global_err_message,  line_no);
}

//------------------------------------------------
static void clear_global_error()
{
	global_errno = 0;
	*global_err_message = 0;
}

//------------------------------------------------
pbms_bool pbms_library_init()
{
	clear_global_error();
	init_count++;
	
	if (init_count == 1 ) {
		CLEAR_SELF();

		CURLcode curl_code = curl_global_init(CURL_GLOBAL_ALL);
		if (curl_code) {
			report_global_error(curl_code, curl_easy_strerror(curl_code) , __LINE__);
			init_count  = 0;
			return 0;
		}
		
		if (! CSThread::startUp()) {
			report_global_error(ENOMEM, "CSThread::startUp() failed.", __LINE__);
			init_count  = 0;
			return 0;
		}
		
		cs_init_memory();
		
		mslib_global_thread = new CSThread( NULL);
		CSThread::setSelf(mslib_global_thread);
		
		CSThread *self = mslib_global_thread;
		if (!mslib_global_thread) {
			report_global_error(ENOMEM, "new CSThread( NULL) failed.", __LINE__);
			init_count  = 0;
			 CSThread::shutDown();
			return 0;
		}
	
		try_(a) {
			pbms_thread_list = new CSThreadList();
		}
		catch_(a);
		report_global_mse_error(self);
		init_count  = 0;
		mslib_global_thread->release();
		mslib_global_thread = NULL;
		cs_exit_memory();
		CSThread::shutDown();
			
		cont_(a);
	}
	
	return(init_count > 0);
}

//------------------------------------------------
void pbms_library_end()
{
	 
	if (init_count == 1 ) {
		init_count  = 0; 
			
		CSThread::setSelf(mslib_global_thread);

		if (pbms_thread_list)
			pbms_thread_list->release();

		
		mslib_global_thread->release();
		mslib_global_thread = NULL;
		cs_exit_memory();
		CSThread::shutDown();
		
		curl_global_cleanup();
	}
	
	if (init_count > 0)
		init_count--;
}

//------------------------------------------------
int pbms_errno(PBMS myhndl)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;

	if (con) {	
		return con->ms_errno;
	}
		
	return global_errno;
}

//------------------------------------------------
const char *pbms_error(PBMS myhndl)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;

	if (con) {	
		return con->ms_err_message;
	}
		
	return global_err_message;
}

//------------------------------------------------
PBMS pbms_connect(const char* host, unsigned int port, const char *database)
{
	PBMS_ConHandle *con = NULL;
	CLEAR_SELF(); 
	
	clear_global_error();

	new_(con, PBMS_ConHandle());
	
	if ((!con) || !CSThread::attach(con)) {
		report_global_error(ENOMEM, "new PBMS_Ref() failed.", __LINE__);
		if (con) {
			con->release();
			con = NULL;
		}
	} else {
		CSThread *self = con;
		
		try_(a) {
				con->ms_initConnection(host, port, database);
				con->ms_ping();
			}
		catch_(a);
			report_global_mse_error(con);
			CSThread::detach(con);
			con = NULL;
			
		cont_(a);
	}
	
	return(con);
}

//------------------------------------------------
PBMS pbms_mysql_connect(SERVER_CON *mysql, const char *database)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	int port;
	char *host = mysql->host;
	PBMS_ConHandle *con = NULL;
	CSThread *self = NULL;
	CLEAR_SELF(); 
	
	clear_global_error();
	
	if (!host) {
		report_global_error(CS_ERR_GENERIC_ERROR, "No host name associated with the MYSQL connection handle." , __LINE__);
		goto done;
	}
	
	if (mysql_query(mysql, "show variables like  \"pbms_port\";")){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_num_rows(results) != 1) {
		report_global_error(CS_ERR_GENERIC_ERROR, "No PBMS blob streaming server associated with the MYSQL connection handle." , __LINE__);
		goto done;
	}
	
	record = mysql_fetch_row(results);
	if ((!record) || !record[1]) {
		report_global_error(CS_ERR_GENERIC_ERROR, "show variables like  \"pbms_port\"; Returned unexpected NULLs." , __LINE__);
		goto done;
	}
	port = atol(record[1]);
	
	new_(con, PBMS_ConHandle());
	
	if ((!con) || !CSThread::attach(con)) {
		report_global_error(ENOMEM, "new PBMS_Ref() failed.", __LINE__);
		if (con) {
			con->release();
			con = NULL;
		}
	}
	
	self = con;
	try_(a) {
		con->ms_initConnection(host, port, database);
		con->ms_ping();
	}
	catch_(a);
	report_global_mse_error(con);
	CSThread::detach(con);
	con = NULL;
	goto done;
		
	cont_(a);
	
	done:	
	if (results)
		mysql_free_result(results);
		
	return(con);
}

//------------------------------------------------
void pbms_close(PBMS myhndl)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
		
	con->ms_setSelf();
	CSThread::detach(con); // This will also release the connection.
}

//------------------------------------------------
pbms_bool pbms_is_blob_reference(PBMS myhndl, const char *ref)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = false;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {
		MSBlobURLRec blob;
		if (ms_parse_blob_url(&blob, (char *)ref)) 
			ok = true;
	}
	
	catch_(a);
	cont_(a);
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_blob_size(PBMS myhndl, const char *ref, size_t *size)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	bool ok = false;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {
		MSBlobURLRec blob;
		if (ms_parse_blob_url(&blob, (char *)ref)) {
			*size = blob.bu_blob_size;
			ok = true;
		} else { // Assume it is a BLOB alias
			CSVector *saved_metadata;
			CSString *data = NULL;
			
			saved_metadata = con->ms_headers.takeHeaders();
			try_(b) {
				con->ms_get_info(ref, true);
				data = con->ms_get_metadata(MS_BLOB_SIZE);
				*size = 	atol(data->getCString());	
				data->release();
				ok = true;
			}
			catch_(b);
			con->report_error(self);
			cont_(b);
			con->ms_headers.setHeaders(saved_metadata);
		}
	}
	
	catch_(a);
	cont_(a);
	return_(ok);
}
/*
 * pbms_add_metadata() and pbms_clear_metadata() deal with metadata for outgoing BLOBs only.
 */
//------------------------------------------------
pbms_bool pbms_add_metadata(PBMS myhndl, const char *name, const char *value)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	bool ok = false;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {
		if (strlen(name) > MS_META_NAME_SIZE)
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Metadata name too long.");
		if (strlen(value) > MS_META_VALUE_SIZE)
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Metadata value too long.");
		
		con->ms_metadata_out.addHeader(name, value);

		ok = true;
	}
	
	catch_(a) {
		con->report_error(self);
	}
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
void pbms_clear_metadata(PBMS myhndl, const char *name)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	
	if (name)
		con->ms_metadata_out.removeHeader(name);
	else
		con->ms_metadata_out.clearHeaders();
}

/*
 * pbms_reset_metadata(), pbms_next_metadata() and pbms_get_metadata_value() deal with metadata for the last
 * BLOB received on the connection.
 */
//------------------------------------------------
int pbms_reset_metadata(PBMS myhndl)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	int count = -1;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {
		count = con->ms_init_fetch();
	}
	
	catch_(a) {
		con->report_error(self);
	}
	
	cont_(a);
	
	return_(count);
}

//------------------------------------------------
pbms_bool pbms_next_metadata(PBMS myhndl, char *name, char *value, size_t *v_size)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	bool ok = false;
	size_t null_size = MS_META_VALUE_SIZE;
	
	con->ms_setSelf();	
	enter_();
	
	if (!v_size)
		v_size = &null_size;
		
	try_(a) {
		const char *m_name, *m_value;
		ok = con->ms_next(&m_name, &m_value);
		if (ok) {
			cs_strcpy(MS_META_NAME_SIZE, name, m_name);
			cs_strcpy(*v_size, value, m_value);

			if (*v_size <= strlen(m_value)) 
				*v_size = strlen(m_value) +1;

		}
	}
	
	catch_(a) {
		con->report_error(self);
	}
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_metadata_value(PBMS myhndl, const char *name, char *buffer, size_t *size)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	bool ok = false;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {
		CSString *data = NULL;
		data = con->ms_get_metadata(name);
		if (data) {
			ok = true;
			cs_strcpy(*size, buffer, data->getCString());
			if (data->length() >= *size) {
				*size = data->length() +1;				
			}
			data->release();
		}
	}
	
	catch_(a) {
		con->report_error(self);
	}
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_md5_digest(PBMS myhndl, char *md5_digest)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	bool ok = false;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {		
		CSString *data = NULL;
		data = con->ms_get_checksum();
		push_(data);
		if (cs_hex_to_bin(16, md5_digest, data->length(), data->getCString()) != 16) {
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Invalid MD5 Digest.");
		} 
		release_(data);
		ok = true;
	}
	
	catch_(a) {
		con->report_error(self);
	}
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_put_data(PBMS myhndl, const char *table, const char *alias, const char *checksum, char *ref, size_t total_data_size, char *data, size_t data_size, PBMS_CALLBACK_FUNC cb, void *caller_data)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = true;
	
	con->ms_setSelf();	

	enter_();
	
	try_(a) {
	
		con->ms_init_put_blob(total_data_size, table, alias, checksum);
		con->set_put_user_data(data, data_size, total_data_size, cb, caller_data);
		con->set_get_user_data(ref, PBMS_BLOB_URL_SIZE -1); // Set things up to receive the BLOB ref back.

		if (curl_easy_perform(con->ms_curl) && !con->ms_throw_error)
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, con->ms_curl_error);
		
		if (con->ms_throw_error)
			throw_();
			
		con->check_reply_status();
	
		ref[con->ms_getDataLen] = 0;
		if (cb && !con->ms_putCBCalled) // The callback needs to be called to notify the caller of completion.
			(cb)(caller_data, data, &data_size, 0);
					
	}
	
	catch_(a);
	con->report_error(self);
	ok = false;
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_data(PBMS myhndl, const char *ref, char *buffer, size_t buffer_size, PBMS_CALLBACK_FUNC cb, void *caller_data)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = true;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {	
		con->set_get_user_data(buffer, buffer_size, cb, caller_data);
		con->ms_init_get_blob(ref, false, false);
		if (curl_easy_perform(con->ms_curl) && !con->ms_throw_error)
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, con->ms_curl_error);
		
		if (con->ms_throw_error)
			throw_();
		
		con->check_reply_status();
			
		if (cb && !con->ms_getCBCalled) // The callback needs to be called to notify the caller of completion.
			(cb)(caller_data, buffer, &buffer_size, 0);
	}
	
	catch_(a);
	con->report_error(self);
	ok = false;
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_info(PBMS myhndl, const char *ref)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = true;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {	
		con->ms_get_info(ref, false);			
	}
	
	catch_(a);
	con->report_error(self);
	ok = false;
	
	cont_(a);
	
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_set_option(PBMS myhndl, enum pbms_option option, const void *in_value)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = true;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {

		switch (option) {
			case PBMS_OPTION_DATABASE: {
				CSString *database = CSString::newString((char *)in_value);
				con->ms_database->release();
				con->ms_database = database;
				break;
			}
				
			case PBMS_OPTION_TRANSMITION_TIMEOUT:
				con->ms_transmition_timeout = *((unsigned int*)in_value);
				break; 
			
			case PBMS_OPTION_HOST:
			case PBMS_OPTION_PORT:
				CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Option is ReadOnly.");
				break;

			default:
				CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Unknown Option.");
		}
	
	}
	
	catch_(a);
	con->report_error(self);
	ok = false;
	cont_(a);
	return_(ok);
}

//------------------------------------------------
pbms_bool pbms_get_option(PBMS myhndl, enum pbms_option option, void *out_value)
{
	PBMS_ConHandle *con = (PBMS_ConHandle*) myhndl;
	pbms_bool ok = true;
	
	con->ms_setSelf();	
	enter_();
	
	try_(a) {

		switch (option) {
			case PBMS_OPTION_DATABASE:
				*((const char**)out_value) = con->ms_database->getCString();
				break;
			
			case PBMS_OPTION_TRANSMITION_TIMEOUT:
				*((unsigned int*)out_value) = con->ms_transmition_timeout;
				break; 
			
			case PBMS_OPTION_HOST:
				*((const char**)out_value) = con->ms_host->getCString();
				break;
				
			case PBMS_OPTION_PORT:
				*((unsigned int*)out_value) = con->ms_port;
				break;

			default:
				CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Unknown Option.");
		}
	
	}
	
	catch_(a);
	con->report_error(self);
	ok = false;
	cont_(a);
	return_(ok);
}


#ifdef UDFS_NO_LONGER_SUPPORTED
#ifdef DRIZZLED

// The drizzle server has internal support for the PBMS engine so all these functions are unneeded:
pbms_bool pbms_init_table(SERVER_CON *mysql __attribute__((unused)), const char *database __attribute__((unused)), const char *table __attribute__((unused)), pbms_bool no_blobs_ok __attribute__((unused))) { return true;}
pbms_bool pbms_reset_table_blobs(SERVER_CON *mysql __attribute__((unused)), const char *database __attribute__((unused)), const char *table __attribute__((unused))){ return true;}
pbms_bool pbms_drop_table_blobs(SERVER_CON *mysql __attribute__((unused)), const char *database __attribute__((unused)), const char *table __attribute__((unused))){ return true;}
pbms_bool pbms_table_renamed(SERVER_CON *mysql __attribute__((unused)), const char *database __attribute__((unused)), const char *old_table __attribute__((unused)), const char *new_table __attribute__((unused))){ return true;}
pbms_bool pbms_enabled(SERVER_CON *mysql __attribute__((unused)), const char *engine __attribute__((unused)), pbms_bool *enabled __attribute__((unused))){ return true;}

#else 

// MySQL servers use triggers to support the PBMS engine:

typedef struct {
	const char *udf_name;
	const char *udf_return_type;
} PBMS_UDFFunction;

static const PBMS_UDFFunction pbms_UDF_Functions[] = { 
	{"pbms_insert_blob_trig",			"STRING"},
	{"pbms_update_blob_trig",			"STRING"},
	{"pbms_delete_blob_trig",			"INTEGER"},
	{"pbms_delete_all_blobs_in_table",	"INTEGER"},
	{"pbms_rename_table_with_blobs",	"INTEGER"},
	{"pbms_enabled_engine",				"INTEGER"},
	{NULL, NULL}
};

typedef struct {
	unsigned short col_pos;
	char	name[PBMS_FIELD_COL_SIZE];
} BlobColRec, *BlobColPtr;

//------------------------------------------------
#define INIT_TRIGGER(t, d, tbl, s) \
	t.append("create trigger "); \
	t.append(d); \
	t.append("."); \
	t.append(tbl); \
	t.append(s); \
	t.append(d); \
	t.append("."); \
	t.append(tbl); \
	t.append(" for each row BEGIN ");

//sprintf(stmt, "set NEW.%s = pbms_insert_blob_trig(\"%s\", \"%s\", %d, NEW.%s);", column, database, table, position, column);
#define SET_INSERT_TRIG(c, d, t, p) \
	insert_trig.append("set NEW."); \
	insert_trig.append(c); \
	insert_trig.append(" = pbms_insert_blob_trig(\""); \
	insert_trig.append(d); \
	insert_trig.append("\", \""); \
	insert_trig.append(t); \
	insert_trig.append("\", "); \
	insert_trig.append(p); \
	insert_trig.append(", NEW."); \
	insert_trig.append(c); \
	insert_trig.append(");"); 	
	
//sprintf(stmt, "set NEW.%s = pbms_update_blob_trig(\"%s\", \"%s\", %d, OLD.%s, NEW.%s);", column, database, table, position, column, column);
#define SET_UPDATE_TRIG(c, d, t, p) \
	update_trig.append("set NEW."); \
	update_trig.append(c); \
	update_trig.append(" = pbms_update_blob_trig(\""); \
	update_trig.append(d); \
	update_trig.append("\", \""); \
	update_trig.append(t); \
	update_trig.append("\", "); \
	update_trig.append(p); \
	update_trig.append(", OLD."); \
	update_trig.append(c); \
	update_trig.append(", NEW."); \
	update_trig.append(c); \
	update_trig.append(");"); 	
	
//sprintf(stmt, "set dummy = pbms_delete_blob_trig(\"%s\", \"%s\", %d, OLD.%s);", database, table, position, column);
#define SET_DELETE_TRIG(c, d, t, p) \
	delete_trig.append("set dummy = pbms_delete_blob_trig(\""); \
	delete_trig.append(d); \
	delete_trig.append("\", \""); \
	delete_trig.append(t); \
	delete_trig.append("\", "); \
	delete_trig.append(p); \
	delete_trig.append(", OLD."); \
	delete_trig.append(c); \
	delete_trig.append(");"); 	

static pbms_bool init_table_for_blobs(SERVER_CON *mysql, const char *database, const char *table, BlobColPtr *info, int *num_blob_columns, bool no_blobs_ok)
{
	pbms_bool ok = false;
	int max_columns, num_columns = 0;
	char stmt[MAX_STMT_SIZE];
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	CSStringBuffer insert_trig, update_trig, delete_trig;
	BlobColPtr iptr = NULL;

	CLEAR_SELF(); 	
	clear_global_error();
	
	// Get a list of all the longblob columns in the table:
	// I would have liked to use information_schema.columns for this but it is not
	// always up to date after a table is dropped and recreated. :(
	
	cs_strcpy(MAX_STMT_SIZE, stmt, "show columns from ");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, ".");
	cs_strcat(MAX_STMT_SIZE, stmt, table);

	if (mysql_query(mysql, stmt)){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	max_columns = mysql_num_rows(results);
	if (!max_columns) {		
		cs_strcpy(MAX_STMT_SIZE, stmt, "No columns found in ");
		cs_strcat(MAX_STMT_SIZE, stmt, database);
		cs_strcat(MAX_STMT_SIZE, stmt, ".");
		cs_strcat(MAX_STMT_SIZE, stmt, table);
		report_global_error(CS_ERR_GENERIC_ERROR, stmt , __LINE__);
		goto done;
	}
	
	iptr = (BlobColPtr) malloc((max_columns) * sizeof(BlobColRec));
	if (!iptr) {
		report_global_error(ENOMEM, "Out of memory" , __LINE__);
		goto done;
	}

	if (info) 
		*info = iptr;
		
	
	for (int i = 0; i < max_columns; i++) {
		char *column, *type;
		record = mysql_fetch_row(results);
		if ((!record) || (!record[0]) || !record[1]) {
			cs_strcat(MAX_STMT_SIZE, stmt, " Returned unexpected NULLs.");
			report_global_error(CS_ERR_GENERIC_ERROR, stmt , __LINE__);
			goto done;
		}
		column = record[0];
		type = record[1];
		
		if (!strcmp(type, "longblob")) {
			iptr[num_columns].col_pos = i+1;
			cs_strcpy(PBMS_FIELD_COL_SIZE, iptr[num_columns].name, column);
			num_columns++;
		}
		
	}
	
	if (num_blob_columns)
		*num_blob_columns = num_columns;
		
	if (!num_columns) {		
		if (no_blobs_ok) 
			ok = true;
		else {
			cs_strcpy(MAX_STMT_SIZE, stmt, "No longblob columns found in ");
			cs_strcat(MAX_STMT_SIZE, stmt, database);
			cs_strcat(MAX_STMT_SIZE, stmt, ".");
			cs_strcat(MAX_STMT_SIZE, stmt, table);
			report_global_error(CS_ERR_GENERIC_ERROR, stmt , __LINE__);
		}
		goto done;
	}

	INIT_TRIGGER(insert_trig, database, table, "_pbms_insert_trig BEFORE INSERT on ")
	INIT_TRIGGER(update_trig, database, table, "_pbms_update_trig BEFORE UPDATE on ")
	INIT_TRIGGER(delete_trig, database, table, "_pbms_delete_trig AFTER DELETE on ")
	
	delete_trig.append("declare dummy integer;"); 

	// For each longblob column install a trigger action:
	for (int i = 0; i < num_columns; i++) {
		char *column = iptr[i].name;
		int position = iptr[i].col_pos;
		
		SET_INSERT_TRIG(column, database, table, position);

		SET_UPDATE_TRIG(column, database, table, position);
		
		SET_DELETE_TRIG(column, database, table, position);
	}
	
	insert_trig.append("END");
	update_trig.append("END");
	delete_trig.append("END");
	
	// Incase the triggers alreay exist, drop them.
	cs_strcpy(MAX_STMT_SIZE, stmt, "drop trigger ");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, ".");
	cs_strcat(MAX_STMT_SIZE, stmt, table);
	cs_strcat(MAX_STMT_SIZE, stmt, "_pbms_insert_trig;");
	mysql_query(mysql, stmt);
	
	cs_strcpy(MAX_STMT_SIZE, stmt, "drop trigger ");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, ".");
	cs_strcat(MAX_STMT_SIZE, stmt, table);
	cs_strcat(MAX_STMT_SIZE, stmt, "_pbms_update_trig;");
	mysql_query(mysql, stmt);
	
	cs_strcpy(MAX_STMT_SIZE, stmt, "drop trigger ");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, ".");
	cs_strcat(MAX_STMT_SIZE, stmt, table);
	cs_strcat(MAX_STMT_SIZE, stmt, "_pbms_delete_trig;");
	mysql_query(mysql, stmt);
	
	// Create the triggers
	if (mysql_query(mysql, insert_trig.getCString())){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_query(mysql, update_trig.getCString())){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_query(mysql, delete_trig.getCString())){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	ok = true;
	
	done:	
	if (results)
		mysql_free_result(results);
		
	if (iptr && ((!ok) || !info)) {
		free(iptr);
		if (info)
			*info = NULL;
	}
		
	return ok;
}

//------------------------------------------------
static pbms_bool check_for_udf(SERVER_CON *mysql, const char *udf, pbms_bool *found)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	char stmt[MAX_STMT_SIZE];
	pbms_bool ok = false;
	
	cs_strcpy(MAX_STMT_SIZE, stmt, "select count(*) from mysql.func where name = \"");
	cs_strcat(MAX_STMT_SIZE, stmt, udf);
	cs_strcat(MAX_STMT_SIZE, stmt, "\";");
	
	if (mysql_query(mysql, stmt)){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		return false;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		return false;
	}

	record = mysql_fetch_row(results);
	if (record) {
		if (!record[0]) {
			cs_strcat(MAX_STMT_SIZE, stmt, " Returned unexpected NULLs.");
			report_global_error(CS_ERR_GENERIC_ERROR, stmt , __LINE__);
			return false;
		}
		*found = (atoi(record[0]) == 1);
		ok = true;
	} else 
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);

	if (results)
		mysql_free_result(results);

	return ok;
}

//------------------------------------------------
static bool load_udfs = true;
pbms_bool pbms_init_udfs(SERVER_CON *mysql)
{
	if (load_udfs) {
		const PBMS_UDFFunction *udf = pbms_UDF_Functions;
		pbms_bool found;
		
		while (udf->udf_name) {
			if (!check_for_udf(mysql, udf->udf_name, &found))
				return false;
			if (!found) {
				char stmt[MAX_STMT_SIZE];
				cs_strcpy(MAX_STMT_SIZE, stmt, "CREATE FUNCTION ");
				cs_strcat(MAX_STMT_SIZE, stmt, udf->udf_name);
				cs_strcat(MAX_STMT_SIZE, stmt, " RETURNS ");
				cs_strcat(MAX_STMT_SIZE, stmt, udf->udf_return_type);
				cs_strcat(MAX_STMT_SIZE, stmt, " SONAME \"libpbms.so\";");
				if (mysql_query(mysql, stmt)){
					report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
					return false;
				}

			}
			udf++;
		}
		load_udfs = false;
	}
	
	return true;
} 

//------------------------------------------------
pbms_bool pbms_init_table(SERVER_CON *mysql, const char *database, const char *table, pbms_bool no_blobs_ok)
{
	CLEAR_SELF(); 	
	clear_global_error();
	
	if (load_udfs && !pbms_init_udfs(mysql)) 
		return false;

	return init_table_for_blobs(mysql, database, table, NULL, NULL, no_blobs_ok);
	
}

//------------------------------------------------
// pbms_reset_table_blobs()
// This function resets all the blobs in a table along with resetting the triggers for the table.
// It should be called after a table has been altered.
pbms_bool pbms_reset_table_blobs(SERVER_CON *mysql, const char *database, const char *table)
{
	pbms_bool ok = false, rename = false;
	int num_columns, cnt;
	char stmt[MAX_STMT_SIZE], tmp_name[20];
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	CSStringBuffer insert_cmd, blob_ref_cmd;
	BlobColPtr blob_columns = NULL;

	CLEAR_SELF(); 	
	clear_global_error();

	if (load_udfs && !pbms_init_udfs(mysql)) 
		return false;
	
	// Reset the triggers on the table.
	if (! init_table_for_blobs(mysql, database, table, &blob_columns, &num_columns, true))
		goto done;
	
	if (!num_columns) {		
		ok = pbms_drop_table_blobs(mysql, database, table);
		goto done;
	}
	
	// Rename any blobs that the table may already have. These will be deleted later if all goes well.
	sprintf(tmp_name, "#%d", time(NULL));
	cs_strcpy(MAX_STMT_SIZE, stmt, "select pbms_rename_table_with_blobs(\"");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, "\", \"");
	cs_strcat(MAX_STMT_SIZE, stmt, table);
	cs_strcat(MAX_STMT_SIZE, stmt, "\", \"");
	cs_strcat(MAX_STMT_SIZE, stmt, tmp_name);
	cs_strcat(MAX_STMT_SIZE, stmt, "\");");
	if (mysql_query(mysql, stmt)){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	record = mysql_fetch_row(results);
	if ((!record) || (!record[0]) || atoi(record[0])) {
		report_global_error(CS_ERR_GENERIC_ERROR, "UDF function pbms_rename_table_with_blobs() reported an error, check the server's stderr output for details.", __LINE__);
		goto done;
	} 
	
	mysql_free_result(results);
	results = NULL;
	rename = true;
	
	// Select all the blobs out of table and register them with PBMS
	insert_cmd.append("select ");
	
	for (num_columns =0; blob_columns[num_columns].col_pos; num_columns++) {
		if (num_columns) insert_cmd.append(", ");
		insert_cmd.append(blob_columns[num_columns].name);
	}
	
	insert_cmd.append(" from ");
	insert_cmd.append(table);
	
	if (mysql_query(mysql, insert_cmd.getCString())){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}

	while  ((record = mysql_fetch_row(results))) {
		blob_ref_cmd.clear();
		blob_ref_cmd.append(" BEGIN  declare dummy TEXT; ");
		cnt = 0;
		for (int j = 0; j < num_columns; j++) {
			size_t len = (record[j])? strlen(record[j]): 0;
			
			if (len && (len < PBMS_BLOB_URL_SIZE)) {
				cnt++;
				//sprintf(stmt,  "set dummy = pbms_insert_blob_trig(\"%s\", \"%s\", %d, \"%s\");", database, table, blob_columns[j].col_pos, record[j]);
				blob_ref_cmd.append("set dummy = pbms_insert_blob_trig(\"");
				blob_ref_cmd.append(database);
				blob_ref_cmd.append("\", \"");
				blob_ref_cmd.append(table);
				blob_ref_cmd.append(", ");
				blob_ref_cmd.append(blob_columns[j].col_pos);
				blob_ref_cmd.append(", \"");
				blob_ref_cmd.append(record[j]);
				blob_ref_cmd.append("\");");
			}
		}
		
		if (cnt) {
			blob_ref_cmd.append(" END ");
			if (mysql_query(mysql, blob_ref_cmd.getCString())){
				report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
				goto done;
			}			
		}
	}
	
	
	// Now delete all the renamed blobs	
	rename = false;
	//sprintf(stmt, "BEGIN  declare dummy integer; dummy = pbms_delete_all_blobs_in_table(\"%s\", \"%s\"); END", database, tmp_name);
	cs_strcpy(MAX_STMT_SIZE, stmt, "select pbms_delete_all_blobs_in_table(\"");
	cs_strcat(MAX_STMT_SIZE, stmt, database);
	cs_strcat(MAX_STMT_SIZE, stmt, "\", \"");
	cs_strcat(MAX_STMT_SIZE, stmt, tmp_name);
	cs_strcat(MAX_STMT_SIZE, stmt, "\");");
	if (mysql_query(mysql, stmt)){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		goto done;
	}
	
	ok = true;
	done:
	if (rename) {
		//sprintf(stmt, "BEGIN  declare dummy integer; dummy = pbms_rename_table_with_blobs(\"%s\", \"%s\", \"%s\"); END", database, tmp_name, table);
		cs_strcpy(MAX_STMT_SIZE, stmt, "select pbms_rename_table_with_blobs(\"");
		cs_strcat(MAX_STMT_SIZE, stmt, database);
		cs_strcat(MAX_STMT_SIZE, stmt, "\", \"");
		cs_strcat(MAX_STMT_SIZE, stmt, tmp_name);
		cs_strcat(MAX_STMT_SIZE, stmt, "\", \"");
		cs_strcat(MAX_STMT_SIZE, stmt, table);
		cs_strcat(MAX_STMT_SIZE, stmt, "\");");
		mysql_query(mysql, stmt);
	}
		
	if (blob_columns)
		free(blob_columns);

	if (results)
		mysql_free_result(results);

	return ok;
}

//------------------------------------------------
static pbms_bool execute_udf(SERVER_CON *mysql, const char *udf_call, int *value)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	char stmt[MAX_STMT_SIZE];
	pbms_bool ok = false;
	
	if (load_udfs && !pbms_init_udfs(mysql)) 
		return false;
	
	cs_strcpy(MAX_STMT_SIZE, stmt, "select ");
	cs_strcat(MAX_STMT_SIZE, stmt, udf_call);
	cs_strcat(MAX_STMT_SIZE, stmt, ";");
	
	if (mysql_query(mysql, stmt)){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		return false;
	}

	results = mysql_store_result(mysql);
	if (!results){
		report_global_error((int) mysql_errno(mysql), (char *) mysql_error(mysql), __LINE__);
		return false;
	}

	record = mysql_fetch_row(results);
	if ((!record) || !record[0]) {
		cs_strcpy(MAX_STMT_SIZE, stmt, "UDF function ");
		cs_strcat(MAX_STMT_SIZE, stmt, udf_call);
		cs_strcat(MAX_STMT_SIZE, stmt, " reported an error, check the server's stderr output for details.");
		report_global_error(CS_ERR_GENERIC_ERROR, stmt, __LINE__);
	} else {
		*value = atoi(record[0]);
		ok = true;
	}
		
	if (results)
		mysql_free_result(results);

	return ok;
}

//------------------------------------------------
pbms_bool pbms_drop_table_blobs(SERVER_CON *mysql, const char *database, const char *table)
{
	char udf_call[MAX_STMT_SIZE];
	int value;
	
	CLEAR_SELF(); 	
	clear_global_error();
	
	cs_strcpy(MAX_STMT_SIZE, udf_call, "pbms_delete_all_blobs_in_table(\"");
	cs_strcat(MAX_STMT_SIZE, udf_call, database);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\", \"");
	cs_strcat(MAX_STMT_SIZE, udf_call, table);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\")");
	
	return (execute_udf(mysql, udf_call, &value) && !value);	
}


//------------------------------------------------
pbms_bool pbms_table_renamed(SERVER_CON *mysql, const char *database, const char *old_table, const char *new_table)
{
	char udf_call[MAX_STMT_SIZE];
	int value;

	CLEAR_SELF(); 	
	clear_global_error();
	
	cs_strcpy(MAX_STMT_SIZE, udf_call, "pbms_rename_table_with_blobs(\"");
	cs_strcat(MAX_STMT_SIZE, udf_call, database);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\", \"");
	cs_strcat(MAX_STMT_SIZE, udf_call, old_table);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\", \"");
	cs_strcat(MAX_STMT_SIZE, udf_call, new_table);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\")");
	
	return (execute_udf(mysql, udf_call, &value) && !value);	
}

//------------------------------------------------
pbms_bool pbms_enabled(SERVER_CON *mysql, const char *engine, pbms_bool *enabled)
{
	char udf_call[MAX_STMT_SIZE];
	int value;

	CLEAR_SELF(); 	
	clear_global_error();
	
	cs_strcpy(MAX_STMT_SIZE, udf_call, "pbms_enabled_engine(\"");
	cs_strcat(MAX_STMT_SIZE, udf_call, engine);
	cs_strcat(MAX_STMT_SIZE, udf_call, "\")");
	
	if (!execute_udf(mysql, udf_call, &value))
		return false;
	
	if (value == -1) {
		cs_strcat(MAX_STMT_SIZE, udf_call, " reported an error, check the server's stderr output for details.");
		report_global_error(CS_ERR_GENERIC_ERROR, udf_call, __LINE__);
		return false;
	}
	
	*enabled = value;
	return true;
}
 #endif // DRIZZLED
 #endif // UDFS_NO_LONGER_SUPPORTED

